package nl.iobyte.themeparkconnector.api.network.io;

import io.socket.client.IO;
import io.socket.client.Socket;
import nl.iobyte.themepark.scheduler.ThemeParkScheduler;
import nl.iobyte.themeparkconnector.ThemeParkConnector;
import nl.iobyte.themeparkconnector.api.client.ClientConnection;
import nl.iobyte.themeparkconnector.api.config.enums.StorageKey;
import nl.iobyte.themeparkconnector.api.message.MessageKey;
import nl.iobyte.themeparkconnector.api.network.drivers.ClientDriver;
import nl.iobyte.themeparkconnector.api.network.drivers.SystemDriver;
import nl.iobyte.themeparkconnector.api.network.objects.CertificateHelper;
import nl.iobyte.themeparkconnector.api.network.objects.SocketDriver;
import nl.iobyte.themeparkconnector.api.packet.objects.AbstractPacket;
import nl.iobyte.themeparkconnector.api.state.states.ConnectingState;
import nl.iobyte.themeparkconnector.api.state.states.IdleState;
import nl.iobyte.themeparkconnector.logger.ThemeParkConnectorLogger;
import okhttp3.OkHttpClient;
import java.net.URISyntaxException;

public class SocketConnector {

    private static final SocketDriver[] drivers = {
            new SystemDriver(),
            new ClientDriver(),
    };

    private Socket socket;

    /**
     * Try to establish connection to the ThemeParkConnector network
     */
    public void setupConnection() {
        if(!ThemeParkConnector.getInstance().getAPI().getStateService().getCurrentState().canConnect())
            return;

        //Preparing socket.io settings
        OkHttpClient okHttpClient = CertificateHelper.ignore(new OkHttpClient.Builder().proxySelector(new NullProxySelector())).build();

        IO.Options opts = new IO.Options();
        opts.callFactory = okHttpClient;
        opts.reconnection = false;
        opts.webSocketFactory = okHttpClient;

        // authentication headers
        opts.query = String.format(
                "type=server&id=%s",
                ThemeParkConnector.getInstance().getAPI().getConfigurationManager().getString(StorageKey.SOCKET_ID)
        );

        //Setup socket.io connection
        try {
            socket = IO.socket(
                    ThemeParkConnector.getInstance().getAPI().getConfigurationManager().getString(StorageKey.SOCKET_URL),
                    opts
            );
        } catch (URISyntaxException e) {
            e.printStackTrace();
        }

        //Register state to be connecting
        ThemeParkConnector.getInstance().getAPI().getStateService().setState(new ConnectingState());

        //Schedule timeout check
        ThemeParkScheduler.runSyncLater(() -> {
            if (!(ThemeParkConnector.getInstance().getAPI().getStateService().getCurrentState() instanceof ConnectingState))
                return;

            cancelWaitingForToken();
            ThemeParkConnectorLogger.toConsole("Connecting timed out.");
            ThemeParkConnector.getInstance().getAPI().getStateService().setState(new IdleState("Connecting to the socket timed out"));
        }, 20 * 35);


        //Register drivers
        for(SocketDriver driver : drivers)
            driver.boot(socket, this);

        socket.connect();
    }

    /**
     * Close connection with Socket
     */
    public void disconnect() {
        if(socket != null)
            socket.disconnect();
    }

    /**
     * Send Packet
     * @param client ClientConnection
     * @param packet AbstractPacket
     */
    public void send(ClientConnection client, AbstractPacket packet) {
        if(client == null || packet == null)
            return;

        if (!ThemeParkConnector.getInstance().getAPI().getStateService().isConnected())
            return;

        String data =  client.getUsedToken() + "@" + ThemeParkConnector.getGson().toJson(packet);
        socket.emit("data", data);
    }

    /**
     * Cancel tokens for all people waiting for one
     */
    public void cancelWaitingForToken() {
        for(ClientConnection client : ThemeParkConnector.getInstance().getAPI().getNetworkingService().getClients()) {
            if(!client.isWaitingForToken())
                continue;

            client.setWaitingForToken(false);
            MessageKey.CLIENT_UNABLE_TO_CONNECT.send(client.getPlayer());
        }
    }

}
